package org.xbib.multiserv.plugin.confserv;

import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import org.xbib.multiserv.api.Command;
import org.xbib.multiserv.api.CommandFunction;
import org.xbib.multiserv.api.CommandServer;
import org.xbib.multiserv.plugin.confserv.util.DefaultCommandClient;
import org.xbib.multiserv.plugin.jdbc.JDBCPlugin;

public class ConfServ extends JDBCPlugin {

    private static final Logger logger = Logger.getLogger(ConfServ.class.getName());

    public ConfServ() {
        super();
    }

    @Override
    public void onStart() throws IOException {
        super.onStart();
        try {
            executeStatement("create table if not exists conf (\"id\" varchar(32), \"type\" varchar(2), \"key\" varchar(256), \"value\" varchar(4000))", Map.of());
            logger.log(Level.INFO, "started");
        } catch (SQLException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
    }

    @Override
    public List<CommandFunction> getFunctions(CommandServer commandServer) {
        List<CommandFunction> list = new ArrayList<>();
        list.add(CommandFunction.of("SLNPGetService", this, SLNPGetService.class));
        list.add(CommandFunction.of("SLNPShowService", this, SLNPShowService.class));
        list.add(CommandFunction.of("SLNPPutService", this, SLNPPutService.class));
        list.add(CommandFunction.of("SLNPDeleteService", this, SLNPDeleteService.class));
        list.add(CommandFunction.of("SLNPAllIsil", this, SLNPAllIsil.class));
        list.add(CommandFunction.of("SLNPSyncConf", this, SLNPSyncConf.class));
        return list;
    }

    @Override
    public void onExecute(CommandServer commandServer,
                          Command<?> command,
                          Map<String, Object> params,
                          List<String> response) {
    }

    public class SLNPGetService extends Command<CommandServer> {

        public SLNPGetService() {
        }

        @Override
        public List<String> execute(Map<String, Object> params) {
            boolean hasSigel = params.containsKey("SigelListe");
            boolean hasISIL = params.containsKey("ISILListe");
            List<String> ids = new ArrayList<>();
            if (hasSigel) {
                ids = Arrays.stream(params.get("SigelListe").toString()
                        .split("[:;]")).map(String::trim).collect(Collectors.toList());
            }
            if (hasISIL) {
                ids = Arrays.stream(params.get("ISILListe").toString()
                        .split("[:;]")).map(String::trim).collect(Collectors.toList());
            }
            if (ids.isEmpty()) {
                return Collections.singletonList("510 NotOkCode:no SigelListe or ISILListe given\n");
            }
            String bstyp2 = (String) params.get("BSTyp2"); // ZFL 4.5
            if (bstyp2 == null) {
                bstyp2 = (String) params.get("BsTyp2"); // ZFL 4.3
            }
            if (bstyp2 == null) {
                bstyp2 = "L"; // assume 'L' does always exist
            }
            List<String> response = new ArrayList<>();
            String errorMessage = null;
            try {
                int i = 0;
                for (String id : ids) {
                    // ZFL 4.3 can send empty isils, ignore it
                    if (id != null && !id.isEmpty()) {
                        Map<String, Object> map= getConf(id, bstyp2);
                        if (map.isEmpty()) {
                            logger.log(Level.WARNING, "no data found for " + id + " with bstyp2 " + bstyp2);
                        }
                        // at position 0: PFL, at position > 0: AFL
                        List<String> list = new ArrayList<>();
                        if (i == 0) {
                            if (hasISIL) {
                                // ISIL, Zweigstelle, Kommunikationsart, IP, Port, ZusatzEmailYN, ZusatzEmailAdr, SSHYN, SSHPort, SSLYN
                                Stream.of("conf.pfl.isil",
                                        "conf.pfl.branchcode",
                                        "conf.pfl.type",
                                        "conf.pfl.ipaddress",
                                        "conf.pfl.port",
                                        "conf.pfl.hasemail",
                                        "conf.pfl.email",
                                        "conf.pfl.hasssh",
                                        "conf.pfl.sshport",
                                        "conf.pfl.hasssl").forEach(k -> list.add(map.containsKey(k) ? map.get(k).toString() : ""));
                                response.add(String.join(",", list));
                            }
                            if (hasSigel) {
                                // Sigel, Zweigstelle, Kommunikationsart, IP, Port, ZusatzEmailYN, ZusatzEmailAdr, SSHYN, SSHPort, SSLYN
                                Stream.of("conf.pfl.sigel",
                                        "conf.pfl.branchcode",
                                        "conf.pfl.type",
                                        "conf.pfl.ipaddress",
                                        "conf.pfl.port",
                                        "conf.pfl.hasemail",
                                        "conf.pfl.email",
                                        "conf.pfl.hasssh",
                                        "conf.pfl.sshport",
                                        "conf.pfl.hasssl").forEach(k -> list.add(map.containsKey(k) ? map.get(k).toString() : ""));
                                response.add(String.join(",", list));
                            }
                        } else {
                            if (hasISIL) {
                                // ISIL, Zweigstelle, Kommunikationsart, IP, Port, ZusatzEmailYN, ZusatzEmailAdr, SSHYN, SSHPort, SSLYN
                                Stream.of("conf.afl.isil",
                                        "conf.afl.branchcode",
                                        "conf.afl.type",
                                        "conf.afl.ipaddress",
                                        "conf.afl.port",
                                        "conf.afl.hasemail",
                                        "conf.afl.email",
                                        "conf.afl.hasssh",
                                        "conf.afl.sshport",
                                        "conf.afl.hasssl").forEach(k -> list.add(map.containsKey(k) ? map.get(k).toString() : ""));
                                response.add(String.join(",", list));
                            }
                            if (hasSigel) {
                                Stream.of("conf.afl.sigel",
                                        "conf.afl.branchcode",
                                        "conf.afl.type",
                                        "conf.afl.ipaddress",
                                        "conf.afl.port",
                                        "conf.afl.hasemail",
                                        "conf.afl.email",
                                        "conf.afl.hasssh",
                                        "conf.afl.sshport",
                                        "conf.afl.hasssl").forEach(k -> list.add(map.containsKey(k) ? map.get(k).toString() : ""));
                                response.add(String.join(",", list));
                            }
                        }
                        i++;
                    }
                }
            } catch (Throwable t) {
                errorMessage = t.getMessage();
                logger.log(Level.SEVERE, t.getMessage(), t);
            }
            return errorMessage != null ?
                    Arrays.asList("510 NotOkCode:" + errorMessage, "") :
                    Arrays.asList(
                            "600 SLNPGetService",
                            "601 LeitwegInfo:" + String.join(":", response),
                            "250 SLNPEndOfData",
                            ""
                    );
        }
    }

    public class SLNPPutService extends Command<CommandServer> {

        public SLNPPutService() {
        }

        @Override
        public List<String> execute(Map<String, Object> params) {
            boolean hasISIL = params.containsKey("ISIL") || params.containsKey("ISILListe");
            String id = null;
            if (hasISIL) {
                id = (String) params.get("ISIL");
                if (id == null) {
                    id = (String) params.get("ISILListe");
                    // cut after ;
                    if (id != null) {
                        int pos = id.indexOf(';');
                        if (pos > 0) {
                            id = id.substring(0, pos);
                        }
                    }
                }
            }
            if (id == null) {
                return Collections.singletonList("510 NotOkCode:no ISIL given\n");
            }
            String bstyp2 = (String) params.get("BSTyp2"); // ZFL 4.5
            if (bstyp2 == null) {
                bstyp2 = (String) params.get("BsTyp2"); // ZFL 4.3
            }
            if (bstyp2 == null) {
                bstyp2 = "L"; // assume 'L' does always exist
            }
            String modus = (String) params.get("Modus");
            if (modus == null) {
                modus = "pfl"; // pfl is default mode
            } else {
                modus = modus.toLowerCase(Locale.ROOT);
            }
            String prefix = "conf." + modus + ".";
            String info = (String) params.get("LeitwegInfo");
            if (info != null) {
                putConf(id, bstyp2, mapInfo(info));
            } else {
                Map<String, Object> mappedParams = params.entrySet().stream()
                        .map(e -> mapEntry(prefix, e))
                        .filter(Objects::nonNull)
                        .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));
                putConf(id, bstyp2, mappedParams);
            }
            return Arrays.asList(
                    "600 SLNPPutService",
                    "510 OK",
                    "250 SLNPEndOfData",
                    ""
            );
        }

    }

    public class SLNPShowService extends Command<CommandServer> {

        public SLNPShowService() {
        }

        @Override
        public List<String> execute(Map<String, Object> params) {
            boolean hasSigel = params.containsKey("Sigel");
            boolean hasISIL = params.containsKey("ISIL");
            String id = null;
            if (hasSigel) {
                id = (String) params.get("Sigel");
            }
            if (hasISIL) {
                id = (String) params.get("ISIL");
            }
            if (id == null) {
                return Collections.singletonList("510 NotOkCode:no Sigel or ISIL given\n");
            }
            String bstyp2 = (String) params.get("BSTyp2"); // ZFL 4.5
            if (bstyp2 == null) {
                bstyp2 = (String) params.get("BsTyp2"); // ZFL 4.3
            }
            if (bstyp2 == null) {
                bstyp2 = "L"; // assume 'L' does always exist
            }
            List<String> info = new ArrayList<>();
            String pfl;
            String zflpfl = "";
            String zflafl = "";
            String errorMessage = null;
            try {
                Map<String, Object> map = getConf(id, bstyp2);
                if (map.isEmpty()) {
                    throw new IllegalArgumentException("no data found for " + id + " with bstyp2 " + bstyp2);
                }
                final List<String> list = new ArrayList<>();
                Stream.of("conf.pfl.type",
                                "conf.pfl.host",
                                "conf.pfl.port")
                        .forEach(k -> list.add(map.containsKey(k) ? map.get(k).toString() : ""));
                pfl = String.join(",", list);
                if (hasISIL) {
                    // ISIL, Zweigstelle, Kommunikationsart, IP, Port, ZusatzEmailYN, ZusatzEmailAdr, SSHYN, SSHPort, SSLYN
                    list.clear();
                    Stream.of("conf.pfl.isil",
                            "conf.pfl.branchcode",
                            "conf.pfl.ipaddress",
                            "conf.pfl.port",
                            "conf.pfl.hasemail",
                            "conf.pfl.email",
                            "conf.pfl.hasssh",
                            "conf.pfl.sshport",
                            "conf.pfl.hasssl").forEach(k -> list.add(map.containsKey(k) ? map.get(k).toString() : ""));
                    zflpfl = String.join(",", list);
                    list.clear();
                    Stream.of("conf.afl.isil",
                            "conf.afl.branchcode",
                            "conf.afl.type",
                            "conf.afl.ipaddress",
                            "conf.afl.port",
                            "conf.afl.hasemail",
                            "conf.afl.email",
                            "conf.afl.hasssh",
                            "conf.afl.sshport",
                            "conf.afl.hasssl").forEach(k -> list.add(map.containsKey(k) ? map.get(k).toString() : ""));
                    zflafl = String.join(",", list);
                }
                if (hasSigel) {
                    // Sigel, Zweigstelle, Kommunikationsart, IP, Port, ZusatzEmailYN, ZusatzEmailAdr, SSHYN, SSHPort, SSLYN
                    list.clear();
                    Stream.of("conf.pfl.sigel",
                            "conf.pfl.branchcode",
                            "conf.pfl.type",
                            "conf.pfl.ipaddress",
                            "conf.pfl.port",
                            "conf.pfl.hasemail",
                            "conf.pfl.email",
                            "conf.pfl.hasssh",
                            "conf.pfl.sshport",
                            "conf.pfl.hasssl").forEach(k -> list.add(map.containsKey(k) ? map.get(k).toString() : ""));
                    zflpfl = String.join(",", list);
                    list.clear();
                    Stream.of("conf.afl.isil",
                            "conf.afl.branchcode",
                            "conf.afl.type",
                            "conf.afl.ipaddress",
                            "conf.afl.port",
                            "conf.afl.hasemail",
                            "conf.afl.email",
                            "conf.afl.hasssh",
                            "conf.afl.sshport",
                            "conf.afl.hasssl").forEach(k -> list.add(map.containsKey(k) ? map.get(k).toString() : ""));
                    zflafl = String.join(",", list);
                }
                info = Arrays.asList(
                        id, // Sigel/ISIL
                        (String) map.getOrDefault("conf.hastanserv", ""), // TanServ
                        (String) map.getOrDefault("conf.getdata", ""), // GetData
                        (String) map.getOrDefault("conf.userdb", ""), // UserDB
                        (String) map.getOrDefault("conf.jason", ""), // JASON
                        pfl, // PFL
                        (String) map.getOrDefault("conf.order", ""), // ORDER
                        "",
                        "",
                        "",
                        "",
                        zflpfl,
                        zflafl,
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        (String) map.getOrDefault("conf.rl", ""), // RL
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        "",
                        ""
                );
            } catch (Throwable t) {
                errorMessage = t.getMessage();
                logger.log(Level.SEVERE, t.getMessage(), t);
            }
            return errorMessage != null ?
                    Arrays.asList("510 NotOkCode:" + errorMessage, "") :
                    Arrays.asList(
                            "600 SLNPShowService",
                            hasSigel ? "602 SigelInfo:" + String.join(";", info) : "602 ISILInfo:" + String.join(";", info),
                            "250 SLNPEndOfData",
                            ""
                    );
        }
    }

    public class SLNPDeleteService extends Command<CommandServer> {

        public SLNPDeleteService() {
        }

        @Override
        public List<String> execute(Map<String, Object> params) {
            boolean hasSigel = params.containsKey("Sigel");
            boolean hasISIL = params.containsKey("ISIL");
            String id = null;
            if (hasSigel) {
                id = (String) params.get("Sigel");
            }
            if (hasISIL) {
                id = (String) params.get("ISIL");
            }
            if (id == null) {
                return Collections.singletonList("510 NotOkCode:no Sigel or ISIL given\n");
            }
            String bstyp2 = (String) params.get("BSTyp2"); // ZFL 4.5
            if (bstyp2 == null) {
                bstyp2 = (String) params.get("BsTyp2"); // ZFL 4.3
            }
            if (bstyp2 == null) {
                bstyp2 = "L"; // assume 'L' does always exist
            }
            try {
                executeStatement("delete from conf c where c.\"id\" = :id and c.\"type\" = :type and c.\"key\" like 'conf.%'",
                        Map.of("id", id, "type", bstyp2));
                return Arrays.asList(
                        "600 SLNPDeleteService",
                        "610 OK",
                        "250 SLNPEndOfData",
                        ""
                );
            } catch (SQLException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                return Arrays.asList(
                        "600 SLNPDeleteService",
                        "510 Error:" + e.getMessage(),
                        "250 SLNPEndOfData",
                        ""
                );
            }
        }
    }
    public class SLNPAllIsil extends Command<CommandServer> {

        public SLNPAllIsil() {
        }

        @Override
        public List<String> execute(Map<String, Object> params) {
            String bstyp2 = (String) params.get("BSTyp2"); // ZFL 4.5
            if (bstyp2 == null) {
                bstyp2 = (String) params.get("BsTyp2"); // ZFL 4.3
            }
            if (bstyp2 == null) {
                bstyp2 = "L"; // assume 'L' does always exist
            }
            try {
                List<Map<String, Object>> list = getRows("select distinct c.\"id\" as \"id\" from conf c where c.\"type\" = :type order by c.\"id\"",
                        Map.of("type", bstyp2));
                List<String> l = list.stream()
                        .map(m -> m.get("id").toString())
                        .toList();
                return Arrays.asList(
                        "600 SLNPAllIsil",
                        "602 ISILInfo:" + String.join(";", l),
                        "250 SLNPEndOfData",
                        ""
                );
            } catch (SQLException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                return Arrays.asList(
                        "600 SLNPAllIsil",
                        "510 Error:" + e.getMessage(),
                        "250 SLNPEndOfData",
                        ""
                );
            }
        }
    }

    public class SLNPSyncConf extends Command<CommandServer> {

        public SLNPSyncConf() {
        }

        @Override
        public List<String> execute(Map<String, Object> params) {
            String bstyp2 = (String) params.get("BSTyp2"); // ZFL 4.5
            if (bstyp2 == null) {
                bstyp2 = (String) params.get("BsTyp2"); // ZFL 4.3
            }
            String host = (String) params.get("Host");
            if (host == null) {
                host = "localhost";
            }
            String portString = (String) params.get("Port");
            if (portString == null) {
                portString = "6142";
            }
            int port = Integer.parseInt(portString);
            DefaultCommandClient client = new DefaultCommandClient();
            try {
                List<Map<String, Object>> list = bstyp2 != null ?
                        getRows("select distinct c.\"id\" as \"id\" from conf c where c.\"type\" = :type order by c.\"id\"",
                                Map.of("type", bstyp2)) :
                        getRows("select distinct c.\"id\" as \"id\" from conf c order by c.\"id\"",
                                Map.of());
                for (String isil : list.stream()
                        .map(m -> m.get("id").toString())
                        .toList()) {
                    StringBuilder sb = new StringBuilder();
                    sb.append("SLNPGetService\n");
                    sb.append("ISILListe:").append(isil).append(";").append(isil).append("\n");
                    if (bstyp2 != null) {
                        sb.append("BSTyp2:").append(bstyp2).append("\n");
                    }
                    sb.append("SLNPEndCommand\n");
                    try {
                        List<String> response = client.sendToPort(host, port, sb.toString());
                        String info = response.stream()
                                .filter(s -> s.startsWith("601 LeitwegInfo:"))
                                .map(s -> s.substring("601 LeitwegInfo:".length()))
                                .collect(Collectors.joining());
                        logger.log(Level.INFO, "isil = " + isil + " leitweginfo = " + mapInfo(info));
                    } catch (IOException e) {
                        logger.log(Level.SEVERE, e.getMessage(), e);
                    }
                }
                return Arrays.asList(
                        "600 SLNPSyncConf",
                        "610 OK:" + list.size(),
                        "250 SLNPEndOfData",
                        ""
                );
            } catch (SQLException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                return Arrays.asList(
                        "600 SLNPSyncConf",
                        "510 Error:" + e.getMessage(),
                        "250 SLNPEndOfData",
                        ""
                );
            }
        }
    }


    private Map<String, Object> getConf(String id, String type) {
        try {
            List<Map<String, Object>> list = getRows("select c.\"key\" as \"key\", c.\"value\" as \"value\" from conf c where c.\"id\" = :id and c.\"type\" = :type and c.\"key\" like 'conf.%' order by c.\"key\"",
                    Map.of("id", id, "type", type));
            return list.stream()
                    .map(m -> Map.of(m.get("key").toString(), m.get("value") != null ? m.get("value") : ""))
                    .flatMap(m -> m.entrySet().stream())
                    .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));
        } catch (SQLException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
        return Map.of();
    }

    private void putConf(String id, String type, Map<String, Object> map) {
        String deleteStatement = "delete from conf where \"id\" = :id and \"type\" = :type and \"key\" = :key";
        List<Map<String, Object>> deleteParams = map.keySet().stream()
                .map(k -> Map.of("id", id, "type", type, "key", (Object) k))
                .toList();
        String insertStatement = "insert into conf (\"id\", \"type\", \"key\", \"value\") values (:id, :type, :key, :value)";
        List<Map<String, Object>> insertParams = map.entrySet().stream()
                .map(e -> Map.of("id", id, "type", type, "key", e.getKey(), "value", e.getValue()))
                .toList();
        try {
            upsertStatement(deleteStatement, deleteParams, insertStatement, insertParams);
        } catch (SQLException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
    }

    private static Map<String, Object> mapInfo(String infoList) {
        Map<String, Object> map = new LinkedHashMap<>();
        String[] info = infoList.split(":"); // colon
        if (info.length > 0) {
            // PFL
            String[] fields = info[0].split(",");
            String prefix = "conf.pfl.";
            if (fields.length > 9) {
                map.put(prefix + "isil", fields[0]);
                map.put(prefix + "sigel", fields[0].replaceAll("DE-", ""));
                map.put(prefix + "branchcode", fields[1]);
                map.put(prefix + "type", fields[2]);
                map.put(prefix + "ipaddress", fields[3]);
                map.put(prefix + "port", fields[4]);
                map.put(prefix + "hasemail", fields[5]);
                map.put(prefix + "email", fields[6]);
                map.put(prefix + "hasssh", fields[7]);
                map.put(prefix + "sshport", fields[8]);
                map.put(prefix + "hasssl", fields[9]);
            }
            if (info.length > 1) {
                // AFL
                fields = info[1].split(",");
                prefix = "conf.afl.";
                if (fields.length > 9) {
                    map.put(prefix + "isil", fields[0]);
                    map.put(prefix + "sigel", fields[0].replaceAll("DE-", ""));
                    map.put(prefix + "branchcode", fields[1]);
                    map.put(prefix + "type", fields[2]);
                    map.put(prefix + "ipaddress", fields[3]);
                    map.put(prefix + "port", fields[4]);
                    map.put(prefix + "hasemail", fields[5]);
                    map.put(prefix + "email", fields[6]);
                    map.put(prefix + "hasssh", fields[7]);
                    map.put(prefix + "sshport", fields[8]);
                    map.put(prefix + "hasssl", fields[9]);
                }
            }
        }
        return map;
    }

    private static Map.Entry<String, Object> mapEntry(String prefix, Map.Entry<String, Object> e) {
        String s = map(e.getKey());
        if (s.isEmpty()) {
            s = map(prefix, e.getKey());
            if (s.isEmpty() || s.equals(prefix)) {
                return null;
            }
        }
        return Map.entry(s, notNull(e.getValue()));
    }

    private static String map(String string) {
        if (string == null) {
            return "";
        }
        return switch (string) {
            case "TanServ" -> "conf.hastanserv";
            case "GetData" -> "conf.getdata";
            case "UserDB" -> "conf.userdb";
            case "ORDER" -> "conf.order";
            case "JASON" -> "conf.jason";
            case "RL" -> "conf.rl";
            default -> "";
        };
    }

    private static String map(String prefix, String string) {
        if (string == null) {
            return "";
        }
        return prefix + (switch (string) {
            case "ISIL" -> "isil";
            case "Sigel" -> "sigel";
            case "KommunikationsArt" -> "type";
            case "Zweigstelle" -> "branchcode";
            case "Host" -> "host";
            case "IP" -> "ipaddress";
            case "Port" -> "port";
            case "ZusatzEmailYN" -> "hasemail";
            case "ZusatzEmailAdr" -> "email";
            case "SSHYN" -> "hasssh";
            case "SSHPort" -> "sshport";
            case "SSLYN" -> "hasssl";
            default -> "";
        });
    }

    private static Object notNull(Object object) {
        return object == null ? "" : object;
    }
}
